#!/usr/bin/perl -w

use strict;

my @distfile = qw(
		  ./std_dip.dist
		  /usr/local/lib/diplomacy/std_dip.dist
		  );
my %dist;
my %abbr;

my $display_only_total = 0;
my $display_only_moves = 0;

my $state = "";
my ($judge, $game, $variant, $phase) = ('', '', '', '');
my %units = ();
my %homes = ();
my %oldhomes = ();
my %rethomes = ();

sub print_units ();
sub print_usage ();
sub read_dist ();
sub init_abbr ();

my $i = 0;

foreach (@ARGV)
{
    $i++;
    
    # Arguments after -- are considered filenames even if starting
    # with a dash.
    last if ($_ eq '--');

    print_usage and exit if ($_ eq '-h');
    
    if ($_ eq '-m')
    {
	$display_only_moves = 1;
	next;
    }

    if ($_ eq '-t')
    {
	$display_only_total = 1;
	next;
    }
    
    print_usage and exit if (/^-/);

    # First non-option
    $i--;
    last;
}

@ARGV = splice @ARGV, $i;

init_abbr;
read_dist;

while (<>)
{
    chomp;    

    if (/^:: Judge: (.*)  Game: (.*)  Variant: (.*)$/)
    {
	die "Judge name changed from $judge to $1\n"
	  if ($judge and $judge ne $1);
	die "Game name changed from $game to $2\n"
	  if ($game and $game ne $2);
	die "Variant changed from $variant to $3\n"
	  if ($variant and $variant ne $3);

	($judge, $game, $variant) = ($1, $2, $3);

	next;
    }
  
    if (/^(S)tarting position for (.)(.*) of (.*?)\./ or
	/^(M)ovement results for (.)(.*) of (.*?)\./)
    {
	%units = ();
	$phase = uc "$2$4$1";

	if ($phase =~ /M$/)
	{
	    %oldhomes = %homes;
	    %homes = ();
	    %rethomes = (); 
	}

	$state = "moves";
	next;
    }

    if (/^(R)etreat orders for (.)(.*) of (.*?)\./)
    {
	%units = ();
	$phase = uc "$2$4$1";
	
	$state = "retreats";
	next;
    }

    if (/^(A)djustment orders for (.)(.*) of (.*?)\./)
    {
	%units = ();
	$phase = uc "$2$4$1";
	
	$state = "adj";
	next;
    }

    next if /^::/;

    if ($state eq "moves" or $state eq "retreats")
    {
	if (/^(\w*):\s*(army|fleet)\s*(.*?)\s*(hold|->|support|convoy|disband)\s*
	    (.*)\.\s*(\(\*(.*)\*\))?/xi or
	    /^(\w*):\s*(army|fleet)\s*(.*?)\./xi
	    )
	{
	    my ($power, $type, $start, $order,
		$rest, $fail) =  map {$_ or ""} ($1, uc $2, lc $3, 
						 lc $4, lc $5, uc $7);
	    $start =~ s/\s*->.*//;

	    $type =~ s/(.).*/$1/;
	    $fail =~ s/.*(dislodged).*/$1/i;

	    if ($fail =~ /dislodged/i)
	    {
		$rethomes{$start} = $oldhomes{$start};
	    }

	    if ($order eq 'disband')
	    {
		# null
	    }
	    elsif ($order eq 'hold' || $order eq 'support' || 
		   $order eq 'convoy')
	    {
		if ($fail !~ /dislodged/i)
		{
		    $homes{$start} = $oldhomes{$start}
		}
		else
		{
		    $rethomes{$start} = $oldhomes{$start};
		}
		
		$units{$power}{$start} = {     order => $order,
					       fail  => $fail,
					       type  => $type
					 };
	    }
	    elsif ($order eq '->')
	    {
		my ($target) = $rest;
		$target =~ s/.*->\s*//;
		
		if (!$fail)
		{
		    if ($state eq "retreats")
		    {
			$homes{$target} = $rethomes{$start};
		    }
		    else
		    {
			$homes{$target} = $oldhomes{$start};
		    }

		    $units{$power}{$target} = {     order => $order,
						    fail  => $fail,
						    type  => $type
					      };
		}
		else
		{
		    next if ($state eq "retreats");

		    if ($fail !~ /dislodged/i)
		    {
			$homes{$start} = $oldhomes{$start}
		    }
		    else
		    {
			$rethomes{$start} = $oldhomes{$start}
		    }

		    $units{$power}{$start} = {     order => $order,
						   fail  => $fail,
						   type  => $type
					     };
		}
	    }
	    else
	    {
		$oldhomes{$start} = $start;
		$homes{$start} = $start;
		$units{$power}{$start} = {     order => "",
					       fail  => $fail,
					       type  => $type
					 };
	    }
	}

	if (/The (.*) in (.*) with no valid retreats was destroyed\./)
	{
	    my $area = $2;

	    foreach my $power (keys %units)
	    {
		if (exists $units{$power}{$area})
		{
		    $units{$power}{$area}{fail} = "DESTROYED";
		}
	    }
	}
    }

    if ($state eq "adj")
    {
	if (/(\w*):\s*(builds|disbands) (a|an) (Army|Fleet) in (.*)\./i)
	{
	    my ($power, $order, $type, $start) =
	      map {$_ or ""} ($1, lc $2, uc $4, lc $5);

	    $type =~ s/(.).*/$1/;

	    if (lc $order eq 'builds')
	    {
		$homes{$start} = $start;
		$oldhomes{$start} = $start;
		$units{$power}{$start} = {     order => "",
					       fail => "",
					       type => $type
					 }
	    }
	    else
	    {
		delete $units{$power}{$start};
	    }
	}
    }

    if (/^The deadline for the first/)
    {
	$state = "";
	print_units;
	next;
    }

    if (/^The next phase/)
    {
	$state = "";
	print_units;
	next;
    }

    if (/^Ownership of supply centers:/)
    {
	$state = "";
	#	print_units;
	next;
    }
    
}

sub print_units ()
{    
    return if (!keys %units);
    return if ($display_only_moves && $phase =~ /[^M]$/);

    print "** $phase **\n\n";

    print sprintf "Country    Unit         Status      Home    By_land By_water\n";
    print '-'x70, "\n";
    foreach my $power (sort keys %units)
    {
	my ($d1, $d2) = (0, 0);
	  
	foreach my $unit (keys %{$units{$power}})
	{
	    my $home =
	      (($units{$power}{$unit}{fail} =~ /dislodged/i or
		$units{$power}{$unit}{fail} =~ /destroyed/i) ?
	       $rethomes{$unit}: $homes{$unit});

	    if (!defined $home)
	    {
		$home = '???';
	    }
	    
	    my $s = sprintf "%-10s %s %-8s ",
	      "$power:", $units{$power}{$unit}{type},
	      uc $abbr{$unit};
	    $s .= sprintf "%-12s ",
 	      ($units{$power}{$unit}{fail} ? 
	       "($units{$power}{$unit}{fail})" : "");
	    $s .= sprintf "(%-6s) ", uc $abbr{$home};

	    #	    print STDERR "$home:$unit:$dist{'F'}{$home}{$unit}\n";

	    my ($ad, $cd, $fd) =
	      ($dist{'A'}{$home}{$unit},
	       $dist{'C'}{$home}{$unit}, 
	       $dist{'F'}{$home}{$unit});

	    $s .= sprintf "%7s %7s",
	      $units{$power}{$unit}{type} eq 'A' ?
	      (defined $ad ? ( ($d1 += $ad) and $ad) : "?", 
	       defined $cd ? ( ($d2 += $cd) and $cd) : "?") :
	      ("",
	       (defined $fd ? ( ($d2 += $fd) and $fd) : "?"));

	    print "$s\n" if (!$display_only_total);
	}
	printf "%-35s %7s %7s %7s\n", "$power:", "Total:", $d1, $d2;
	  
	print "\n" if (!$display_only_total);
    }
    
    print "\n" if ($display_only_total);
}

sub read_dist ()
{
    my $open = 0;
    my ($t, $area, $carea, $num);

    foreach (@distfile)
    {
	$open = 1;
	last if open DIST, $_;
	$open = 0;
    }

    die "Couldn't open any of the following files for reading distance data:\n",
      join "  \n", @distfile if !$open;

    print STDERR "Reading distance data:\n";

    while (<DIST>)
    {
	chomp;

	if (/^-\[(.*)\]/)
	{
	    $t = $1;
	    $t =~ s/(.).*/$1/;
	    next;
	}

	if (s/^(\d+)\s+(.*),.*/$2/)
	{
	    $num = $1;
	    if ($num == 0)
	    {
		$carea = lc $2;
		if ($t ne 'F')
		{
		    $carea =~ s/\s*\(.*\)//;
		}
		else
		{
		    $carea =~ s/\(sc\)/(south coast)/;
		    $carea =~ s/\(nc\)/(north coast)/;
		    $carea =~ s/\(ec\)/(east coast)/;
		}	    
	    }
	}

	foreach $area (split /\s*,\s*/, $_)
	{
	    if ($t ne 'F')
	    {
		$area =~ s/\s*\(.*\)//;
	    }
	    else
	    {
		$area =~ s/\(sc\)/(south coast)/;
		$area =~ s/\(nc\)/(north coast)/;
		$area =~ s/\(ec\)/(east coast)/;
	    }

	    # Kludge to fix problem with multi-coast areas.
	    if (!exists $dist{$t}{$carea}{lc $area})
	    {
		$dist{$t}{$carea}{lc $area} = $num;
	    }
	}
    }    
}

sub init_abbr ()
{
    $abbr{'???'} = "???";
    $abbr{'adriatic sea'} = "adr";
    $abbr{'aegean sea'} = "aeg";
    $abbr{albania} = "alb";
    $abbr{ankara} = "ank";
    $abbr{apulia} = "apu";
    $abbr{armenia} = "arm";
    $abbr{'baltic sea'} = "bal";
    $abbr{'barents sea'} = "bar";
    $abbr{belgium} = "bel";
    $abbr{berlin} = "ber";
    $abbr{'black sea'} = "bla";
    $abbr{bohemia} = "boh";
    $abbr{'gulf of bothnia'} = "bot";
    $abbr{brest} = "bre";
    $abbr{budapest} = "bud";
    $abbr{bulgaria} = "bul";
    $abbr{'bulgaria (east coast)'} = "bul/ec";
    $abbr{'bulgaria (south coast)'} = "bul/sc";
    $abbr{burgundy} = "bur";
    $abbr{clyde} = "cly";
    $abbr{constantinople} = "con";
    $abbr{denmark} = "den";
    $abbr{'eastern mediterranean'} = "eas";
    $abbr{edinburgh} = "edi";
    $abbr{'english channel'} = "eng";
    $abbr{finland} = "fin";
    $abbr{galicia} = "gal";
    $abbr{gascony} = "gas";
    $abbr{greece} = "gre";
    $abbr{'helgoland bight'} = "hel";
    $abbr{holland} = "hol";
    $abbr{'ionian sea'} = "ion";
    $abbr{'irish sea'} = "iri";
    $abbr{kiel} = "kie";
    $abbr{london} = "lon";
    $abbr{livonia} = "lvn";
    $abbr{liverpool} = "lvp";
    $abbr{'gulf of lyon'} = "lyo";
    $abbr{'mid-atlantic ocean'} = "mao";
    $abbr{marseilles} = "mar";
    $abbr{moscow} = "mos";
    $abbr{munich} = "mun";
    $abbr{'north africa'} = "naf";
    $abbr{'north atlantic ocean'} = "nao";
    $abbr{naples} = "nap";
    $abbr{norway} = "nor";
    $abbr{'north sea'} = "nth";
    $abbr{'norwegian sea'} = "nwg";
    $abbr{paris} = "par";
    $abbr{picardy} = "pic";
    $abbr{piedmont} = "pie";
    $abbr{portugal} = "por";
    $abbr{prussia} = "pru";
    $abbr{rome} = "rom";
    $abbr{ruhr} = "ruh";
    $abbr{rumania} = "rum";
    $abbr{serbia} = "ser";
    $abbr{sevastopol} = "sev";
    $abbr{silesia} = "sil";
    $abbr{skagerrak} = "ska";
    $abbr{smyrna} = "smy";
    $abbr{spain} = "spa";
    $abbr{'spain (north coast)'} = "spa/nc";
    $abbr{'spain (south coast)'} = "spa/sc";
    $abbr{'st petersburg'} = "stp";
    $abbr{'st petersburg (north coast)'} = "stp/nc";
    $abbr{'st petersburg (south coast)'} = "stp/sc";
    $abbr{sweden} = "swe";
    $abbr{syria} = "syr";
    $abbr{trieste} = "tri";
    $abbr{tunis} = "tun";
    $abbr{tuscany} = "tus";
    $abbr{tyrolia} = "tyr";
    $abbr{'tyrrhenian sea'} = "tys";
    $abbr{ukraine} = "ukr";
    $abbr{venice} = "ven";
    $abbr{vienna} = "vie";
    $abbr{wales} = "wal";
    $abbr{warsaw} = "war";
    $abbr{'western mediterranean'} = "wes";
    $abbr{yorkshire} = "yor";
    $abbr{switzerland} = "swi";
}

sub print_usage ()
{
    print <<END;
Usage: dipunithome.perl [options] [filenames]

    Options:
      --   Consider all arguments after this to be filenames.
      -h   Show this text.
      -m   Display statistics only after movement phases.
      -t   Display only the totals for each power.
END
      return 1;    
}

